import test_data from "./test_data.ts";

export type INode = {
    kind: INodeKind;
    filename: string;
    path: string;
    id: number;
};

type INodeKind =
    | { type: "dir"; children: INode[] }
    | { type: "file"; url: string };

type GithubContentEntry = {
    type: string;
    name: string;
    download_url: string;
};

async function githubTree(name: string, ref: string): Promise<INode> {
    return {
        id: 0,
        path: "",
        filename: name,
        kind: {
            type: "dir",
            children: await githubListRecursive(name, ref),
        },
    };
}

async function githubListRecursive(
    name: string,
    ref: string,
    path = "",
    nextId = [1],
): Promise<INode[]> {
    const fileNodes: GithubContentEntry[] = await fetch(
        `https://api.github.com/repos/${name}/contents${path}?ref=${ref}`,
    ).then((res) => res.json());
    return Promise.all(
        fileNodes.map(async (node): Promise<INode> => {
            const id = nextId[0]++;
            if (node.type === "dir") {
                return await githubListRecursive(
                    name,
                    ref,
                    `${path}/${node.name}`,
                    nextId,
                ).then((children): INode => ({
                    id,
                    filename: node.name,
                    path,
                    kind: { type: "dir", children },
                }));
            } else if (node.type === "file") {
                return {
                    id,
                    filename: node.name,
                    path,
                    kind: { type: "file", url: node.download_url },
                };
            }
            throw new Error();
        }),
    );
}

function generateTreeHtml(node: INode): string {
    return `<ul>${generateNodeHtml(node)}</ul>`;
}

function generateNodeHtml(node: INode): string {
    if (node.kind.type === "dir") {
        const children = node.kind.children
            .map((node) => generateNodeHtml(node))
            .join("");
        return `<li>
            <input type="checkbox" id="checkbox-${node.id}" checked>
            ${node.filename}/
            <ul>${children}</ul>
        </li>`;
    } else if (node.kind.type === "file") {
        return `
            <li>
                <input type="checkbox" id="checkbox-${node.id}" checked>
                ${node.filename}
            </li>
        `;
    }
    throw new Error();
}

type HtmlTree = Map<number, HtmlINode>;

type HtmlINode = {
    node: INode;
    parentId?: number;
    checkbox: HTMLInputElement;
};

function queryHtmlINodes(tree: INode): HtmlTree {
    const nodes = new Map<number, HtmlINode>();
    queryHtmlINodesRecursive(nodes, tree);
    return nodes;
}

function queryHtmlINodesRecursive(
    nodes: HtmlTree,
    node: INode,
    parentId?: number,
) {
    const checkbox = document
        .querySelector<HTMLInputElement>(`#checkbox-${node.id}`)!;
    if (node.kind.type === "dir") {
        nodes.set(node.id, { node, parentId, checkbox });
        for (const child of node.kind.children) {
            queryHtmlINodesRecursive(nodes, child, node.id);
        }
    } else if (node.kind.type === "file") {
        nodes.set(node.id, { node, parentId, checkbox });
    }
}

function hydrateHtmlTree(tree: HtmlTree) {
    for (const node of tree.values()) {
        node.checkbox.addEventListener("change", () => {
            if (node.node.kind.type === "dir") {
                setCheckChildrenRecursively(
                    tree,
                    node.node.id,
                    node.checkbox.checked,
                );
            }
            checkParentDirCheckRecursively(tree, node.parentId);
        });
    }
}

function setCheckChildrenRecursively(
    tree: HtmlTree,
    id: number,
    state: boolean,
) {
    const node = tree.get(id)!;
    if (node.node.kind.type !== "dir") {
        throw new Error();
    }
    for (const { id: childId } of node.node.kind.children) {
        const child = tree.get(childId)!;
        child.checkbox.checked = state;
        if (child.node.kind.type === "dir") {
            setCheckChildrenRecursively(tree, childId, state);
        }
    }
}

function checkParentDirCheckRecursively(tree: HtmlTree, parentId?: number) {
    if (parentId === undefined) {
        return;
    }
    const parent = tree.get(parentId);
    if (parent === undefined) {
        throw new Error();
    }
    if (parent.node.kind.type !== "dir") {
        return;
    }
    const checked = parent.node.kind.children
        .some(({ id }) => tree.get(id)!.checkbox.checked);
    parent.checkbox.checked = checked;
    checkParentDirCheckRecursively(tree, parent.parentId);
}

type FileMatch = {
    filename: string;
    path: string;
    linetext: string;
    linenr: number;
};

async function* searchTree(
    tree: HtmlTree,
    pattern: RegExp,
): AsyncGenerator<FileMatch> {
    for (const node of tree.values()) {
        if (node.node.kind.type !== "file") {
            continue;
        }
        if (!node.checkbox.checked) {
            continue;
        }
        const text = await fetch(node.node.kind.url)
            .then((res) => res.text());
        const lines = text.split("\n").map((v, i) => [v, i] as const);
        for (const [linetext, i] of lines) {
            if (!pattern.test(linetext)) {
                continue;
            }
            const { filename, path } = node.node;
            const linenr = i + 1;
            yield { filename, path, linetext, linenr };
        }
    }
}

const repoNameInput = document
    .querySelector<HTMLInputElement>("#repo-name")!;
const repoRefInput = document
    .querySelector<HTMLInputElement>("#repo-ref")!;
const gitProviderSelect = document
    .querySelector<HTMLSelectElement>("#git-provider")!;
const loadRepoButton = document
    .querySelector<HTMLButtonElement>("#load-repo")!;
const fileTreeDiv = document
    .querySelector<HTMLDivElement>("#file-tree")!;
const searchPatternInput = document
    .querySelector<HTMLInputElement>("#search-pattern")!;
const searchButton = document
    .querySelector<HTMLButtonElement>("#search")!;
const searchStatusSpan = document
    .querySelector<HTMLSpanElement>("#search-status")!;
const searchResultsDiv = document
    .querySelector<HTMLDivElement>("#search-results")!;

loadRepoButton.onclick = async () => {
    const name = repoNameInput.value;
    const ref = repoRefInput.value;
    const provider = gitProviderSelect.value;

    if (provider === "github") {
        const tree = await githubTree(name, ref);
        fileTreeDiv.innerHTML = generateTreeHtml(tree);
        const htmlTree = queryHtmlINodes(tree);
        hydrateHtmlTree(htmlTree);

        searchStatusSpan.innerText = ``;
        searchButton.onclick = async () => {
            const patternText = searchPatternInput.value;
            const pattern = new RegExp(patternText);
            searchStatusSpan.innerText = "Searching ...";
            searchResultsDiv.innerHTML = "";
            let matches = 0;
            for await (const match of searchTree(htmlTree, pattern)) {
                matches += 1;
                searchResultsDiv.innerHTML += `
                    <div>
                        <p>${match.path}/${match.filename} line ${match.linenr}</p>
                        <code>${match.linetext}</code>
                    </div>
                `;
            }
            if (matches === 0) {
                searchStatusSpan.innerText = "** Nothing found **";
            } else {
                searchStatusSpan.innerText = `⋆˙⟡ Done, ${matches} matches ⟡˙⋆`;
            }
        };
    }
};

searchButton.onclick = () => {
    searchStatusSpan.innerText = `No repository loaded :(`;
};
